@@ -10,7 +10,8 @@
 #include "iw.h"
 
 static int offchannel(struct nl80211_state *state, struct nl_cb *cb,
-		      struct nl_msg *msg, int argc, char **argv)
+		      struct nl_msg *msg, int argc, char **argv,
+		      enum id_input id)
 {
 	char *end;
 
