@@ -671,8 +671,8 @@ void print_vht_info(__u32 capa, const __u8 *mcs)
 	PRINT_VHT_CAPA(22, "+HTC-VHT");
 	/* max A-MPDU */
 	/* VHT link adaptation */
-	PRINT_VHT_CAPA(29, "RX antenna pattern consistency");
-	PRINT_VHT_CAPA(30, "TX antenna pattern consistency");
+	PRINT_VHT_CAPA(28, "RX antenna pattern consistency");
+	PRINT_VHT_CAPA(29, "TX antenna pattern consistency");
 
 	printf("\t\tVHT RX MCS set:\n");
 	tmp = mcs[0] | (mcs[1] << 8);
@@ -702,3 +702,16 @@ void print_vht_info(__u32 capa, const __u8 *mcs)
 	tmp = mcs[6] | (mcs[7] << 8);
 	printf("\t\tVHT TX highest supported: %d Mbps\n", tmp & 0x1fff);
 }
+
+void iw_hexdump(const char *prefix, const __u8 *buf, size_t size)
+{
+	int i;
+
+	printf("%s: ", prefix);
+	for (i = 0; i < size; i++) {
+		if (i && i % 16 == 0)
+			printf("\n%s: ", prefix);
+		printf("%02x ", buf[i]);
+	}
+	printf("\n\n");
+}
